-- syntax sugar:
--    case (foo) {
--        ['case1'] = 'stuff',
--        ['case2'] = 'more',
--        [3]       = 'even more',
--        default   = 'bar'
--    }
function case(x)
	return function(cases)
		local r = cases[x] or cases.default
		return type(r) == "function" and r() or r
	end
end

function string.count(s, pattern)
	local count = 0
	for _ in s:gmatch(pattern) do
		count = count + 1
	end
	return count
end

local input = {}

function joystickDown(index, btn, ...)
	if btn then
		return love.joystick.isDown(index,btn) or joystickDown(index, ...)
	end
	return false
end

function input.jump()
	return joystickDown(0,0,1,2,3,4,5,6,7,8) or love.keyboard.isDown('up')
end

function input.move()
	if love.keyboard.isDown('left') then return -1 end
	if love.keyboard.isDown('right') then return 1 end
	if math.abs(love.joystick.getAxis(0,0)) > .5 then
		return love.joystick.getAxis(0,0)
	end
	if math.abs(love.joystick.getAxis(0,2)) > .5 then
		return love.joystick.getAxis(0,2)
	end
	return galse
end

Timer   = require 'timer'
states  = require 'states'
HC      = require 'hardoncollider'
vector  = require 'hardoncollider.vector'
Polygon = require 'hardoncollider.polygon'
require 'AnAL'

shadow_color = {30,20,10,255}

function love.load()
	Timer.add(1, function() states.switch(states.start) end)
	HC.init(100, function(dt, a,b, dx,dy)
		local player, obstacle
		if a._groups.obstacle then
			player, obstacle = b,a
			dx,dy = -dx,-dy
		elseif b._groups.obstacle then
			player, obstacle = b,a
		else
			return
		end

		player:move(dx, dy)

		local v = vector(dx,dy)
		local mirrored = player.vel:projectOn(v) - 2 * v
		player.vel = player.vel - mirrored
		v:normalize_inplace()
		player.onGround = player.onGround or (v.y < -.7) -- consider anything up to 45 deg as ground
		player.push = v * 100
	end, function(_, a,b)
		local player
		if     a._groups.player then player = a
		elseif b._groups.player then player = b
		else return end
		player.onGround = false
	end)

	objects = {
		ground = HC.addRectangle(0,550, 800,500),
		-- borders
		wall_left_1  = HC.addRectangle(0,400, 20,150),
		wall_left_2  = HC.addPolygon(20,400, 0,0, 0,400),
		wall_right_1 = HC.addRectangle(800-20,400, 20,150),
		wall_right_2 = HC.addPolygon(800-20,400, 800,0, 800,400),

		-- stairs
		stair_1 = HC.addRectangle(700,490, 50,20),
		stair_2 = HC.addRectangle(600,440, 50,20),
		stair_3 = HC.addRectangle(500,390, 50,20),

		-- platfoms
		hanging_wall    = HC.addRectangle(200,400, 20,100),
		platform_center = HC.addRectangle(300,300, 150,20),
		platform_high1  = HC.addRectangle(400,200, 100,20),
		platform_high2  = HC.addRectangle(550,150, 80,20),
		platform_high3  = HC.addRectangle(400,100, 80,20),
		platform_high4  = HC.addRectangle(150,150, 100,20),
	}
	objects.stair_1:rotate(math.pi/13)
	objects.stair_2:rotate(math.pi/11)
	objects.stair_3:rotate(math.pi/9)
	objects.hanging_wall:rotate(math.pi/7)
	objects.platform_center:rotate(math.pi/5)
	objects.platform_high1:rotate(-.1)
	objects.platform_high2:rotate( .1)
	objects.platform_high3:rotate( .1)
	objects.platform_high4:rotate(-.1)

	shadow_casters = {}
	for _,obj in pairs(objects) do
		HC.addToGroup("obstacle", obj)
		if obj ~= objects.ground and obj ~= objects.wall_left_1 and obj ~= objects.wall_left_2 and obj ~= objects.wall_right_1 and obj ~= objects.wall_right_2 then
			shadow_casters[#shadow_casters+1] = obj._polygon
		end
	end

	player = HC.addCircle(400, 400, 15)
	player.anim = newAnimation(love.graphics.newImage('fire.png'), 32,64, .05, 6)
	player.vel = vector(0,0)
	player.floater = Timer.Oscillator(1, function(s)
		local phi = s * 2 * math.pi
		local x = math.sin(phi * 2) * 1.5
		local y = math.sin(phi) * 2
		return x,y, math.sin(phi) / 15
	end)
	HC.addToGroup("player", player)
	function player:update(dt)
		self.anim:update(dt)
		self.acc = vector(0,1500)
		if self.onGround then
			local friction = .3
			local move = input.move()
			if input.jump() then
				self.vel.y = -600
				friction = .1
			elseif move then
				self.vel.x = self.vel.x + 40 * move
				if move < 0 then
					self.vel.x = math.max(self.vel.x, -200)
				else
					self.vel.x = math.min(self.vel.x, 200)
				end
				friction = .1
			end
			self.vel = self.vel * (1 - friction)
		else
			if love.joystick.getAxis(0,0) < -.5 or love.keyboard.isDown('left') then
				self.vel.x = math.max(self.vel.x - 5, -200)
			elseif love.joystick.getAxis(0,0) > .5 or love.keyboard.isDown('right') then
				self.vel.x = math.min(self.vel.x + 5, 200)
			end
		end
		self.vel = self.vel + self.acc * dt
		self:move((self.vel*dt):unpack())
	end

	function player:draw(s)
		love.graphics.setColor(255,255,255)
		local x,y = self:center()
		local dx,dy,angle = self.floater(love.timer.getDelta())
		self.anim:draw(x+dx,y+dy, angle,s^3,s^3, 16,47)
	end

	mask_radius = 150
	light_mask = love.image.newImageData(2*mask_radius,2*mask_radius)
	light_mask:mapPixel(function(x,y)
		x = (x - mask_radius) / mask_radius
		y = (y - mask_radius) / mask_radius
		local d = math.sqrt(x * x + y * y)
		return 255,255,255, math.min(1, math.pow(d,1.5)) * 255
	end)
	light_mask = love.graphics.newImage(light_mask)
	love.graphics.setBackgroundColor(80,60,40)

	love.mouse.setVisible(false)
end

t = 0
function love.update(dt)
	local dt = math.min(dt, 1 / 30)
	t = t + dt
	Timer.update(dt)
	states.update(dt)
	HC.update(dt)
	player:update(dt)
end

function love.keypressed(key)
	case (key) {
		['d'] = function() _do_debug = not _do_debug end,
		['q'] = function() love.event.push('q') end,
	}
end

function love.draw()
	love.graphics.setColor(255,255,255)
	local ppos = vector(player:center())
	ppos.x = ppos.x + math.sin(2 * math.pi * t) * 1.5
	ppos.y = ppos.y + math.sin(4 * math.pi * t) * 2

	-- cast shadows
	love.graphics.setColor(shadow_color)
	for _,poly in ipairs(shadow_casters) do
		if ppos:dist(poly.centroid) < 1.5 * mask_radius then
			local startv, endv = poly.vertices[1], poly.vertices[1]
			local is_last_hidden = false
			local vertices = {}
			for i,v in ipairs(poly.vertices) do
				local w = poly.vertices[(i % #poly.vertices) + 1]
				local c = .5 * (w + v)
				local n = (w - v):perpendicular()
				local is_hidden = n * (c - ppos) < 0
				if is_hidden then
					if not is_last_hidden then
						startv = v
					end
				elseif is_last_hidden then
					endv = v
				end
				is_last_hidden = is_hidden
			end

			vertices[1] = startv.x
			vertices[2] = startv.y
			vertices[3] = endv.x
			vertices[4] = endv.y

			-- translate last and first vertex away from player
			endv = endv + (endv - ppos) * 800
			startv = startv + (startv - ppos) * 800
			vertices[5] = endv.x
			vertices[6] = endv.y
			vertices[7] = startv.x
			vertices[8] = startv.y
			love.graphics.polygon('fill', unpack(vertices))
		end
	end

	-- draw objects
	love.graphics.setColor(160,100,30)
	for _, obj in pairs(objects) do
		obj:draw('fill')
	end

	-- draw mask
	love.graphics.setColor(shadow_color)
	local s = math.random() * .06 + .97
	local left, right = ppos.x - s * mask_radius, ppos.x + s * mask_radius
	local top, bottom = ppos.y - s * mask_radius, ppos.y + s * mask_radius
	love.graphics.draw(light_mask, ppos.x,ppos.y, 0,s,s, mask_radius,mask_radius)
	love.graphics.rectangle('fill', 0,0,      left,love.graphics.getHeight())
	love.graphics.rectangle('fill', right,0,  love.graphics.getWidth(),love.graphics.getHeight())
	love.graphics.rectangle('fill', 0,0,      love.graphics.getWidth(),top)
	love.graphics.rectangle('fill', 0,bottom, love.graphics.getWidth(),love.graphics.getHeight())
	player:draw(s)

	states:draw()

	love.graphics.setColor(200,200,200)
	--if player.push then
	--	local v = vector(player:center())
	--	local w = v + player.push
	--	love.graphics.setColor(200,0,0)
	--	love.graphics.line(v.x, v.y, w.x, w.y)
	--	w = v + player.vel
	--	love.graphics.setColor(200,200,0)
	--	love.graphics.line(v.x, v.y, w.x, w.y)
	--end
	if _do_debug then
		love.graphics.printf('FPS: '..love.timer.getFPS(), 10, 12, 400)
	end
end
