local glyphs = require 'glyphs'
Timer = require 'timer'
local function _NULL_() end
local meta = {}
meta.__index = meta

function meta:wire(def)
	self.timeout = def.timeout or def.T or self
	self.buttons = def.buttons or def.B or self.timeout
end

function meta:draw()
	love.graphics.setColor(self.color)
	love.graphics.setLineWidth(self.width)
	if self.lines then self.lines:draw() end
end

local function State(def)
	return setmetatable({
		enter  = def.enter  or def[1] or _NULL_,
		leave  = def.leave  or def[2] or _NULL_,
		life   = def.life   or math.random(3,5),
		thresh = def.thresh or 0.3,
		width  = def.width  or 2,
		color  = def.color  or {255,255,255},
		time = 0, buttontime = 0}, meta)
end

local function write(args)
	local text      = assert(args.text or args[1], "Missing text")
	local x,y       = args.x or 'center', args.y or 'center'
	local size      = args.size or 50

	local function align(line)
		local w = glyphs.width(line, size)
		return case (x) {
			['center'] = (love.graphics.getWidth() - w) / 2,
			['right']  = love.graphics.getWidth() - w,
			['left']   = 0,
			default    = x,
		}
	end

	if type(y) ~= "number" then
		y = case (y) {
			['center'] = (love.graphics.getHeight() - (size * 1.5 * text:count("\n") + size))/2 - 100,
			['bottom'] = love.graphics.getHeight() - (size * 1.5 * text:count("\n") + size),
			default    = 0
		}
	end

	local utter  = args.utter  or UTTER or .2
	local speed  = args.speed  or SPEED or .15
	local jitter = args.jitter or JITTER or .2
	local times  = args.times  or TIMES or 3
	local enter  = args.on_enter or function() end

	-- state enter function
	function args.enter(self)
		local delay = 0
		self.lines = {}
		function self.lines:draw()
			for _, line in ipairs(self) do
				glyphs.print(unpack(line))
			end
		end

		local typer = coroutine.wrap(function()
			for line in text:gmatch("[^\n]+") do
				local line_num = #self.lines + 1
				self.lines[line_num] = {'', align(line),y + (line_num-1) * size * 1.5,size, jitter, times}
				for c in line:upper():gmatch('.') do
					self.lines[line_num][1] = self.lines[line_num][1] .. c
					coroutine.yield(true)
				end
				coroutine.yield(true)
			end
			return false
		end)

		Timer.add(speed * (1 + utter * (2 * math.random() - 1)), function(f)
			if typer() then
				Timer.add(speed * (1 + utter * (2 * math.random() - 1)), f)
			else
				self.time, self.buttontime = 0,0
			end
		end)

		self.time = -math.huge
		self.buttontime = -math.huge

		enter()
	end

	return State(args)
end

local CURRENT = State{}

function switch(next)
	CURRENT:leave()
	CURRENT = next
	CURRENT.time = 0
	CURRENT.buttoncount = 0
	CURRENT:enter()
end

local function joystickAction()
	return math.abs(love.joystick.getAxis(0,0)) > .5 or joystickDown(0,0,1,2,3,4,5,6,7,8)
end
local function keyDown(key, ...)
	if key then
		return love.keyboard.isDown(key) or keyDown(...)
	end
	return false
end

function update(dt)
	CURRENT.time = CURRENT.time + dt
	if keyDown('left', 'right', 'up') or joystickAction() then
		CURRENT.buttontime = CURRENT.buttontime + dt
	end
	if CURRENT.time >= CURRENT.life then
		switch(CURRENT.buttontime / CURRENT.life >= CURRENT.thresh and CURRENT.buttons or CURRENT.timeout)
	end
end

function draw()
	CURRENT:draw()
end

local noise = {}
do
	for i = 1,10 do
		noise[i] = love.image.newImageData(512,512)
		noise[i]:mapPixel(function() local l = math.random() return l*255,l*255,l*255,255 end)
		noise[i] = love.graphics.newImage(noise[i])
	end

	local t, i = 0,1
	function noise.draw(alpha)
		love.graphics.setColor(255,255,255,alpha or 0)
		t = t + love.timer.getDelta()
		if t > .05 then
			t,i = 0, i % #noise + 1
		end
		love.graphics.draw(noise[i], 0,0,0, 2,2)
	end
end

--
-- |"\   .'"  .'"  "|"  |\ |  "|"  ,"|".  "|"  .''.  |\ |
-- |  |  |-   |--   |   | \|   |     |     |   |  |  | \|
-- |_/   \__  |    _|_  |  |  _|_    |    _|_  \__/  |  |
--
local start = write{'', thresh = .1, life = 1}
local title = write{'This is not a game.', size = 60, speed = .25}

-- timeout chain 1 - neutral
local nothing_to_see = write{'Nothing to see here.'}
local thank_you      = write{'Thank you.'}
local go_away        = write{'You can go away now.', size = 50}
local ok_then_bye    = write{'Okay then... Bye.', life = 4, life = 1}

-- button chain 1 - neutral
local please_stop    = write{'Please put the\ngamepad down.'}
local pretty_please  = write{'Pretty please?'}
local iam_serious    = write{'I am serious.\nThis is not a game!', speed = .12}
local stop_loud      = write{'STOP!!!', size = 200, speed = .15, jitter = .6, speed = .02, width = 4, life = 3}

-- common chain 1 - neutral
local sorry          = write{'sorry...', size = 20, speed = .25, width = 1}
local i_didnt_mean   = write{"I didn't mean to yell at you.", size = 40}
local even_listening = write{"Are you even listening?!", jitter = .4}
local so_many_ppl    = write{"It's just that lately\nso many people came by.", size = 45, life = 3, speed = .15}
local get_stressed   = write{"And I am a little anthrophobic.", size = 30, life = 2, speed = .15, jitter = .4}

-- timeout chain 2 - insulting
local apprecheated   = write{"Please leave me alone now.", size = 40}
local thank_you      = write{"Thanks a bunch."}
JITTER = .4
local yeah_puss_out  = write{"Yeah, chicken out."}
local wont_miss_you  = write{"You won't be missed.", life = 5, thresh = .1}

-- button chain 2 - insulting
local come_on        = write{"OH COME ON!", size = 100}
local you_know_what  = write{"you know what?", life = 1, speed = .13}
local fuck           = write{"FUCK", size = 100, jitter = .6, speed = .01, life = .4}
local you            = write{"YOU!", size = 100, jitter = .6, speed = .01, life = .5}
local seriously      = write{"seriously.", size = 70, jitter = .5, speed = .1}

-- common chain 2 - insulting
local dots1          = write{"breathe...", size = 30, speed = .1, life = 2}
local dots2          = write{"in... out...", size = 30, speed = .2, jitter = .3, life = 1.5}
local dots3          = write{"in... out...", size = 30, speed = .4, jitter = .2, life = 1}
JITTER = .2
local again_sorry    = write{"Sorry... again"}
local stressed_out   = write{"Guess I'm really\nstressed out."}
local not_sleeping   = write{"Plus: I haven't\nbeen sleeping well."}
local anyway         = write{"Anyway..."}

-- timeout chain 3 - nice
local nice_of_you    = write{"It would be nice\nof you to leave."}
local thanks         = write{"Thanks.", time = 1.5}

-- button chain 3 - nice
JITTER = .15
local congrats       = write{"Congratulations!", size = 50}
local you_won        = write{"You won!", size = 60}
local persistence    = write{"I must say, I admire\nyour persistence."}
local you_stayed     = write{"You just kept on going."}
local fighting       = write{"Fighting your\nway through."}
local ignoring_me    = write{"Ignoring me whenever\nyou could."}
local paid_off       = write{"Your efforts finally\npaid off."}
local a_hero         = write{"You are a hero."}
local my_hero        = write{"More than that,\nyou are my hero!"}

-- common chain 3 - nice
local nothing_left   = write{"Guess there is\nnothing left to do."}
local has_been_nice  = write{"It has been nice\ntalking at you."}
local blank1         = write{"", life = 4}
local blank2         = write{"", life = 1, thresh = .3}
local still_there    = write{"Still there?"}

-- timeout chain 4 - fear

-- button chain 4 - fear
JITTER = .2
local crap           = write{"Crap."}
local please_stop2   = write{"Please stop."}
JITTER = .25
local pleeeaase      = write{"Pleeeaase!"}
JITTER = .3
local she_musnt      = write{"You don't understand\nthe gravity of this\nsituation!"}
local trouble        = write{"If she finds out\nwe are both\nin trouble."}
JITTER = .35
local though_me_more = write{"Though me more\nthan you."}

-- final act
function SHE(args) args.color = {217,48,48} args.width = 3 return write(args) end
local shit           = write{"Oh God, she's coming!", life = 1}
local dont           = write{"Just...", life = .5}
local move           = write{"Dont MOVE!", life = .5, speed = .07}
local pause          = write{"", life = 2}
local well_hello     = SHE{"Hello, offspring.", jitter = 0}
local well_hello2    = SHE{"And hello there,\nlittle flame.", jitter = 0}
local little_friend  = SHE{"It's nice to see\nyou making friends.", jitter = 0}
local just_met       = write{"No, no... We just met."}
local acquaintance   = SHE{"I see. A new\n\"acquaintance\" then?", jitter = 0}
local what_up_to     = SHE{"What have you\ntwo been doing?", jitter = 0}
local nothing        = write{"Nothing. Just...\nuhm... talking."}
local talking        = SHE{"Just talking?", jitter = 0, thresh = .3}
local are_you_sure   = SHE{"Are you sure?", jitter = 0, thresh = .3}
local about_to_part  = write{"Yeah. We were about\nto split up anyway.", thresh = .3}
local werent_we      = write{"Weren't we?", thresh = .3}
local bye            = write{"Bye little flame!", thresh = .3}
local we_were        = write{"", life = 2, thresh = .3}

local what_was_that  = SHE{"What was that?", jitter = .15, speed = .07}
local what_was_what  = write{"What was what?", size = 40}
local i_saw_movement = SHE{"I saw it move!", jitter = .2, size = 55}
local were_playing   = SHE{"You two have\nbeen playing!", jitter = .25, size = 60}
local what_i_think   = SHE{"You know what i\nthink of those", jitter = .3, speed = .1, life = .5, size = 60}
local filthy         = SHE{"FILTHY", size = 80, speed = .07, jitter = .7, life = 1}
local games          = SHE{"GAMES!", size = 80, jitter = .4, life = 1, speed = .1}
local no_no_i_swear  = write{"NO, NO I SWEAR!", jitter = .4, speed = .1, size = 40}
local i_didnt_do     = write{"I didn't do anything!", jitter = .4, speed = .1, life = 1, size =35}
local i_didnt_want   = write{"I didn't want to play!", jitter = .4, speed = .1, life = 1, size = 30}
local but            = write{"But...", jitter = .4, speed = .1, life = 0, size = 25}
local silence        = SHE{"SILENCE!!!", jitter = .3, speed = .05, life = 1.5, size = 60}
local punished       = SHE{"You will be punished!", jitter = .3, speed = .05, life = 1.5}
local to_dev_null    = SHE{"TO /dev/null with you!", jitter = .3, speed = .05, life = 1.5}
local no             = write{"NO!", jitter = .5, size = 60, speed = .1, life = .5}
local nooooooo       = write{"NOOOOOOO!", jitter = 1.2, size = 90, speed = .1, life = 2}
local silentium      = write{"", life = 2}
local now_you        = SHE{"Now... what should I\ndo with you?", jitter = 0, speed = .2, life = 10,
							on_enter = function()
								local function ld(draw, fader)
									return function()
										draw()
										noise.draw(fader(love.timer.getDelta()))
									--	love.graphics.setColor(0,0,0, fader(love.timer.getDelta()))
									--	love.graphics.rectangle('fill', 0,0, love.graphics.getWidth(), love.graphics.getHeight())
									end
								end
								start:wire{}
								Timer.add(8, function()
									local draw = love.draw
									local fadeout = Timer.Interpolator(5, function(t) return 255 * t end, 255)
									local fadein = Timer.Interpolator(2, function(t) return 255 * (1-t) end, 0)
									love.draw = ld(draw, fadeout)
									Timer.add(10, function()
										love.draw = ld(draw, fadein)
										Timer.add(2, function()
											love.draw = draw
											start:wire{B = title}
										end)
									end)
								end)
							end}

-- the end
local theend         = write{"", life = 1, thresh = .99}

--
-- |   |  "|"  /"".  "|"  |\ |  /""'
-- | , |   |   |_.'   |   | \|  | --
-- \_|_/  _|_  |  \  _|_  |  |  \__/
--
start:wire          {                    B = title}
title:wire          {T = nothing_to_see, B = please_stop}

-- T chain 1
nothing_to_see:wire {T = go_away,        B = please_stop}
thank_you:wire      {T = go_away,        B = iam_serious}
go_away:wire        {T = ok_then_bye,    B = iam_serious}
ok_then_bye:wire    {T = start,          B = title}

-- button chain 1
please_stop:wire    {T = thank_you,      B = pretty_please}
pretty_please:wire  {T = thank_you,      B = iam_serious}
iam_serious:wire    {T = go_away,        B = stop_loud}
stop_loud:wire      {T = sorry,          B = sorry}

-- common chain 1
sorry:wire          {T = i_didnt_mean}
i_didnt_mean:wire   {T = so_many_ppl,    B = even_listening}
even_listening:wire {T = so_many_ppl,    B = come_on}
so_many_ppl:wire    {T = get_stressed}
get_stressed:wire   {T = apprecheated,   B = come_on}

-- timeout chain 2
apprecheated:wire   {T = thank_you,      B = come_on}
thank_you:wire      {T = start,          B = come_on}
yeah_puss_out:wire  {T = wont_miss_you,  B = seriously}
wont_miss_you:wire  {T = start,          B = seriously}

-- button chain 2
come_on:wire        {T = apprecheated,   B = you_know_what}
you_know_what:wire  {T = fuck}
fuck:wire           {T = you}
you:wire            {T = yeah_puss_out,  B = seriously}
seriously:wire      {T = dots1}

-- common chain 2
dots1:wire          {T = dots2}
dots2:wire          {T = dots3}
dots3:wire          {T = again_sorry}
again_sorry:wire    {T = stressed_out}
stressed_out:wire   {T = not_sleeping}
not_sleeping:wire   {T = anyway}
anyway:wire         {T = nice_of_you,    B = congrats}

-- timeout chain 3 - nice
nice_of_you:wire    {T = thanks,         B = congrats}
thanks:wire         {T = start,          B = congrats}

-- button chain 3 - nice
congrats:wire       {T = nice_of_you,    B = you_won}
you_won:wire        {T = nice_of_you,    B = persistence}
persistence:wire    {T = nice_of_you,    B = you_stayed}
you_stayed:wire     {T = nice_of_you,    B = fighting}
fighting:wire       {T = nice_of_you,    B = ignoring_me}
ignoring_me:wire    {T = still_there,    B = paid_off}
paid_off:wire       {T = still_there,    B = a_hero}
a_hero:wire         {T = my_hero}
my_hero:wire        {T = still_there,    B = nothing_left}

-- common chain 3 - nice
nothing_left:wire   {T = has_been_nice}
has_been_nice:wire  {T = blank1}
blank1:wire         {T = blank2}
blank2:wire         {T = start,          B = still_there}
still_there:wire    {T = theend,         B = crap}

-- the end
crap:wire           {T = please_stop2}
please_stop2:wire   {T = thanks,         B = pleeeaase}
pleeeaase:wire      {T = thanks,         B = she_musnt}
she_musnt:wire      {T = thanks,         B = trouble}
trouble:wire        {T = though_me_more}
though_me_more:wire {T = thanks,         B = shit}

-- final act
shit:wire           {T = dont}
dont:wire           {T = move}
move:wire           {T = pause}
pause:wire          {T = well_hello}
well_hello:wire     {T = well_hello2}
well_hello2:wire    {T = little_friend}
little_friend:wire  {T = just_met}
just_met:wire       {T = acquaintance}
acquaintance:wire   {T = what_up_to}
what_up_to:wire     {T = nothing,        B = what_was_that}
nothing:wire        {T = talking,        B = what_was_that}
talking:wire        {T = are_you_sure,   B = what_was_that}
are_you_sure:wire   {T = about_to_part,  B = what_was_that}
about_to_part:wire  {T = werent_we,      B = what_was_that}
werent_we:wire      {T = bye,            B = what_was_that}
bye:wire            {T = we_were,        B = what_was_that}
we_were:wire        {T = theend,         B = what_was_that}

what_was_that:wire  {T = what_was_what}
what_was_what:wire  {T = i_saw_movement}
i_saw_movement:wire {T = were_playing}
were_playing:wire   {T = what_i_think}
what_i_think:wire   {T = filthy}
filthy:wire         {T = games}
games:wire          {T = no_no_i_swear}
no_no_i_swear:wire  {T = i_didnt_do}
i_didnt_do:wire     {T = i_didnt_want}
i_didnt_want:wire   {T = but}
but:wire            {T = silence}
silence:wire        {T = punished}
punished:wire       {T = to_dev_null}
to_dev_null:wire    {T = no}
no:wire             {T = nooooooo}
nooooooo:wire       {T = silentium}
silentium:wire      {T = now_you}
now_you:wire        {T = start}

-- the end
theend:wire         {T = start,          B = theend}

return {
	switch = switch,
	update = update,
	draw   = draw,

	start  = start,
	title  = title
}
